local Bomb = {}

local assets = require('utils.assets')
local audio = require('utils.audio')
local colors = require('utils.colors')
local filters = require('utils.filters')
local signal = require('lib.signal')
local vec = require('lib.vector-light')
local drawShadow = require('utils.shadow')
local Explosion = require('entities.explosion')

local game = nil

local GRAVITY = 1000
local TIME = 3.0
local IGNITE_T = 0.25
local UNPICKABLE_T = 0.8

function Bomb.setup(_game)
  game = _game
  game.register('Bomb', Bomb)
end

local mt = {
  is_pickable = true, -- will need it?
  is_bomb = true
}
mt.__index = mt

function mt:ignite()
  if not self.is_ignited then
    self.is_ignited = true
    self.t = math.min(self.t or IGNITE_T, IGNITE_T)
  end
end

function mt:put(x, y, putter)
  self.idle = false
  self.x, self.y, self.z = x-4, y-5, 12
  self.vx, self.vy = 0, 0
  self.vz = 300
  -- self.t = (self.t or TIME) - 0.01

  if not self._added then
    self._added = true
    game.safeAdd(self)
  end
end

function mt:throw(x, y, z, dir_x, dir_y, vx, vy)
  self.idle = false
  self.x, self.y, self.z = x-6, y-5, z
  -- self.t = (self.t or TIME) - 0.01


  local ndx, ndy = vec.normalize(dir_x, dir_y)

  self.vx, self.vy = vx + ndx * 80, vy + ndy * 50 -- y less because map is smaller vertically
  self.vz = 200

  if not self._added then
    self._added = true
    game.safeAdd(self)
  end
end

function mt:getY()
  return -- hack
    self.y + (self.z >= 0 and 16 or 0)
end

function mt:getOrigin()
  return self.x + 5, self.y + 5
end

function mt:getRect()
  return self.x, self.y, 10, 10 -- ?
end

function mt:tryFall()
  local rx, ry, rw, rh = self:getRect()
  local _, len = game.world:queryRect(rx+3,ry+3,rw-6,rh-6, filters.queryTile)
  if len == 0 then -- no tile!
    self.is_falling = true -- irreversable
  end
  return len == 0
end

function mt:setUnpickableBy(player, time)
  self._unpickable_by = player
  self._unpickable_by_t = time
end

function mt:update(dt)
  if self.is_falling then
    self.is_pickable = false -- so it cannot be picked anymore
    self.vz = self.vz - GRAVITY * dt

    self.x = self.x + self.vx * dt
    self.y = self.y + self.vy * dt
    self.z = self.z + self.vz * dt

  else
    if not self.idle then
      self.x = self.x + self.vx * dt
      self.y = self.y + self.vy * dt
      self.z = self.z + self.vz * dt

      self.vz = self.vz - GRAVITY * dt
    else
      if not self.t then self.t = TIME end -- only idle can explode, no mid-air explosions!
      self.t = self.t - dt

      if self.t < 0 then
        self:explode()
      end
    end

    if self.vx ~= 0 then
      self.last_dir = self.vx > 0 and 1 or -1
    end

    self.is_pickable = false
    if self.z <= 1 then
      if self:tryFall() then return end
      self.z = 0
      self.vx = self.vx * 0.6
      self.vy = self.vy * 0.6
      self.vz = -self.vz * 0.6



      if math.abs(self.vz) > 3 then
        audio.playRandom('bomb_jump')
      else
        self.is_pickable = not not self.t
      end
    end

    if math.abs(self.vx) < 1 and math.abs(self.vy) < 1 and math.abs(self.vz) < 1 and self.z < 1 then
      self.z = 0
      self.is_pickable = not not self.t
      self.idle = true
    end


    if self.z < 5 and math.abs(self.vz) < 5 then
      if self:tryFall() then return end
      self.idle = true
      self.z = self.z * 0.5
      self.vz = self.vz * 0.5
    end

    if self.z < 10 then
      local new_x, new_y, cols, len = game.world:move(self, self.x, self.y, filters.moveBomb)
      for i, col in ipairs(cols) do
        local ox, oy = col.other:getOrigin()
        local bx, by = self:getOrigin()
        local dx, dy = vec.normalize(bx-ox, by-oy)
        self:push(dx*100, dy*100, 0)
      end
      self.x, self.y = new_x, new_y
    else
      game.world:update(self, self.x, self.y)
    end
  end

  if self._unpickable_by_t then
    self._unpickable_by_t = self._unpickable_by_t - dt
    if self._unpickable_by_t <= 0 then
      self._unpickable_by = nil
      self._unpickable_by_t = nil
    end
  end
end

function mt:carriedUpdate(dt)
end

function mt:carriedDraw(x, y, dir)
  love.graphics.setColor(colors.white)
  local dir = (dir == -1) and -1 or 1
  local dy = -26
  local dx = (dir == -1) and 13 or -4
  assets.draw('bomb_basic', x+dx, y+dy, 0, dir, 1)
  self:drawNum(x, y+dy+4, 0)
end

function mt:drawNum(bx, by, bz)
  if self.t and not self.is_falling then
    local n = ('%02d'):format((self.t*10/TIME)*10)
    -- love.graphics.setColor(0, 0, 0)
    -- love.graphics.print(n, self.x+2.5, self.y+4, 0, 0.5, 0.5)
    love.graphics.setColor(255, 255, 255)
    love.graphics.print(n, bx+1, by+3-bz, 0, 0.45, 0.45)
  end
end

function mt:push(vx, vy, vz)
  self.idle = false
  self.z = 1
  self.vx, self.vy, self.vz = vx, vy, vz
end

function mt:canBePickedBy(player)
  return
    ((not self._unpickable_by) or (player ~= self._unpickable_by)) and (self.z >= 0) and (self.z < 10)
end

function mt:bePicked()
  audio.playRandom('pick')
  self.is_pickable = false
  if self._added then
    game.remove(self)
    self._added = false
  end
end

function mt:draw()
  if not self.is_falling then drawShadow(self.x-1, self.y+8, (0.95-(self.z/100))^2) end

  love.graphics.setColor(colors.white)

  local dx = (self.last_dir == -1) and 13 or -4
  assets.draw('bomb_basic', self.x+dx, self.y-self.z-4, 0, self.last_dir, 1)
  self:drawNum(self.x, self.y, self.z)

  -- love.graphics.rectangle('line', self:getRect())
end

function mt:explode()
  if self._added then
    game.remove(self)
    self._added = false
  end
  -- signal.emit('explosion', self.x+5, self.y+5) -- should go to explosion?
  game.add(Explosion.new(self.x+5, self.y+5, self))
end

function Bomb.new(x, y)
  return setmetatable({
    is_pickable = false, -- you need to throw
    x = x, y = y, z = 0,
    push_vx, push_vy = 0, 0,
  }, mt)
end

return Bomb
