-- boxes does not give anything to players, the map is responsible for that!

local assets = require('utils.assets')
local colors = require('utils.colors')
local audio = require('utils.audio')
local filters = require('utils.filters')
local signal = require('lib.signal')
local drawShadow = require('utils.shadow')
local Explosion = require('entities.explosion')
local box_particles = require('particles.box_particles')

local game = nil

local REMOVE_T = 1

local Box = {}
local mt = {
  is_box = true,
  is_pickable = true -- tmp
}
mt.__index = mt

function mt:ignite() -- there is a bomb inside after all
  if self.fall_t or (not self.is_pickable) then return end

  if not self.is_ignited then
    self.is_ignited = true
    self.is_picked = true -- ?
    self.is_pickable = false
  end
end

function mt:isPicked()
  return self.is_picked
end

function mt:onBePicked()
  self.is_picked = true
  self.is_pickable = false
  self.psystem:moveTo(self.x+5, self.y+5)
  self.psystem:emit(10)
  audio.playRandom('pick')
end

function mt:getOrigin()
  return self.x + 5, self.y + 6
end

function mt:getRect()
  return self.x, self.y, 11, 12
end

function mt:getY()
  local ft = self.fall_t or 0
  return self.y + (ft > 0.05 and 0 or 25)
end

function mt:explode()
  if not self:canExplode() then return end

  self.is_exploded = true
  game.remove(self) -- no need to really worry about dust as explosion will give its own
  local ox, oy = self:getOrigin()
  local e = Explosion.new(ox, oy, self)
  game.add(e)
  self.psystem:moveTo(self.x+5, self.y+5)
  self.psystem:emit(10)
  signal.emit('box:exploded')
end

function mt:update(dt)
  self.t = self.t + dt

  if self.is_ignited then
    self.ignited_t = (self.ignited_t or love.math.random()*0.2+0.1) - dt
    if self.ignited_t < 0 then
      self:explode()
    end
    return --[[ !!!! ]]
  end

  if self.is_pickable == false and not self.is_picked and self.t > 4 then
    self.is_pickable = true

    signal.emit('stomp', nil, 0, self:getOrigin())
    self.psystem:emit(10)
    audio.playRandom('stomp')

  end
  if self.is_picked then
    self.picked_t = (self.picked_t or 0) + dt
  end
  self.psystem:update(dt)

  -- fall
  if self.fall_t then
    self.fall_t = self.fall_t + dt
    self.vz = (self.vz or 0) - 900 * dt
    self.z = (self.z or 0) + self.vz * dt
  else
    local tx, ty, tw, th = self:getRect()
    local _, len = game.world:queryRect(tx+3, ty+3, tw-6, th-6, filters.queryTile)

    if len == 0 then
      self.fall_t = 0
    end
  end

  if self.fall_t or self.is_picked then
    self.remove_t = (self.remove_t or REMOVE_T) - dt -- some time for box or dust to fall
    if self.remove_t < 0 then
      game.remove(self)
    end
  end
end

function mt:canTouchBomb()
  if self.fall_t then return false end -- not anymore
  if self.t < 3.25 then return false end -- not yet (still in the air)
  if self.is_exploded then return false end -- already did
  return true
end

function mt:canExplode()
  if self.fall_t then return false end -- not anymore
  if self.t < 4 then return false end -- not yet
  if self.is_exploded then return false end -- already did
  return true
end

function mt:draw()
  love.graphics.setColor(255, 255, 255)

  if self.fall_t then -- fall falling
    assets.draw('box', self.x, self.y -2 - (self.z or 0))

  elseif self.t < 4 then -- intro falling
    local t = (self.t-2) / 2
    -- local n = 1-t

    if t > 0 then
      local z = 300 * (1-(t^8))
      assets.draw('box', self.x, self.y -2 - z)
    end

    drawShadow(self.x, self.y+8, t)
  else -- ready
    if self.is_picked then
      love.graphics.draw(self.psystem)
      if self.picked_t < 0.06 then
        game.toggleBlink(true)
        assets.draw('box', self.x, self.y - 2)
        game.toggleBlink(false)
      end
    else
      assets.draw('box', self.x, self.y - 2)
    end
  end
  love.graphics.setColor(255, 255, 255)
end

function Box.setup(_game)
  game = _game
  game.register('Box', Box)
end

function Box.new(x, y)
  signal.emit('box_spawned')
  return setmetatable({
    t = 0,
    x = x, y = y,
    is_pickable = false, -- it's not pickable before it falls!
    picked_t = 0,
    psystem = box_particles.new()
  }, mt)
end

return Box
