local animset = require('utils.animset')
local assets = require('utils.assets')
local later = require('utils.later')
local filters = require('utils.filters')
local audio = require('utils.audio')
local colors = require('utils.colors')
local vec = require('lib.vector-light')
local signal = require('lib.signal')
local ism = require('lib.ism')
local drawShadow = require('utils.shadow')
local map = require('map')
local input = require('input')

local tex = assets.img('tex.png')
local player_dust = require('particles.player_dust')

local ACCELERATION = 750
local SQUASH_TIME = 0.1
local INIT_JUMP_V = 250
local INTRO_GRAVITY = 600
local RISE_GRAVITY = 900
local RISE_LOW_GRAVITY = RISE_GRAVITY * 2.5
local STOMP_GRAVITY = RISE_GRAVITY * 10
local FALL_GRAVITY = RISE_GRAVITY * 2
local FAIL_FALL_GRAVITY = FALL_GRAVITY
local FAIL_FALL_T = 1

local game = nil
local newPlayerAnimset = nil
local Player = {}

local mt = {
  is_player = true
}
mt.__index = mt


function mt:init(id, x, y)
  self.animset = newPlayerAnimset('fall')
  self.ism = ism.new(self, 'intro') -- ...
  self.dir_x, self.dir_y = 1, 0
  self:setDir(1, 0)
  self.blink_t = 0
  self.can_be_controlled = true --?

  -- particles
  self.dust_psystem = player_dust.new()

  -- self.current_anim = 'idle'

  self.z = 0
  self.vx, self.vy, self.vz = 0, 0, 0
  self.ax, self.ay = 0, 0
  self.color = self:getColor()
  -- self.total_t = 0
  -- self.init_random = love.math.random()
  -- self.blinting_till_t = -1

  self.push_vx, self.push_vy = 0, 0 -- separate from vx,vy because they need defferent deacceleration

  self.payload = nil -- if not nil then the player is carrying something
end


function mt:update(dt)
  self.ism:update(dt)
  self.animset:update(dt)
  self.dust_psystem:update(dt)

  if self.vx ~= 0 or self.vy ~= 0 then
    self.x = self.x + self.vx * dt
    self.y = self.y + self.vy * dt
  end

  if self.payload and self.payload.carriedUpdate then
    self.payload:carriedUpdate(dt)
  end

  self.blink_t = self.blink_t - dt
end


function mt:draw()
  if self.z >= 0 then
    drawShadow(self.x-1, self.y+2, (1-(self.z/1000))^2)
    love.graphics.setColor(self.color)
  end

  love.graphics.setColor(colors.white)
  love.graphics.draw(self.dust_psystem)

  love.graphics.setColor(self.color)
  local blink = (self.blink_t > 0 and (math.floor(self.blink_t * 10) % 2 == 0))
  if blink then game.toggleBlink(true) end
  self.animset:draw(self.x-3, self.y-11-self.z, self.anim_dir_x)
  if blink then game.toggleBlink(false) end

  if self.payload and self.payload.carriedDraw then
    local offsets = Player.payload_offset[self.animset:getState()] or Player.payload_offset['default']
    local oy = offsets[self.animset:getFrame()] or 0
    self.payload:carriedDraw(self.x, self.y + oy - self.z, self.anim_dir_x)
  end


  love.graphics.setColor(colors.white)
  -- love.graphics.rectangle('line', self:getRect())
end

--------------- INTRO STATE ---------------

function mt:onIntroStart()
  local safe_x, safe_y = map.getSafeSpotForPlayer(self)

  if not safe_x then
    self.ism:set('limbo') -- map is over
    self.limbo = true
  else
    game.world:update(self, safe_x, safe_y)
    self.x, self.y = safe_x, safe_y
  end

  self.payload = nil

  self.vx, self.vy, self.vz = 0, 0, 0
  self.z = 500
end

function mt:onIntroUpdate(dt)
  if self.limbo then return end
  self:playAnim('fall')
  self.z = self.z + self.vz * dt
  self.vz = self.vz - (INTRO_GRAVITY) * dt
  if self.z <= 0 then
    self.z = 0
    self.ism:set('idle')
    self.blink_t = 1

    self:tryFailFall()
  end
end

function mt:onIntroEnd()
  -- stomp without damages -- ?
  if self:hasGroundUnderneath() then
    self:emitDust('stomp')
    signal.emit('stomp', self.id, 0, self:getOrigin())
    audio.playRandom('stomp')
    audio.playRandom('intro_fall')
  else
    self:tryFailFall()
  end
end

--------------- DEATH STATE ---------------

function mt:onDeathStart(dt)
  self.vy = 140
end
function mt:onDeathUpdate(dt)
  self:playAnim(self.vy > 0 and 'rise' or 'fall')
  self.vy = self.vy - 700 * dt
  self.y = self.y - self.vy * dt * 5
end

--------------- FAIL FALL STATE ---------------

function mt:onFailFallUpdate(dt)
  self.vz = self.vz - FAIL_FALL_GRAVITY * dt
  self.z = self.z + self.vz * dt

  self.vx, self.vy = self.vx * 0.8, self.vy * 0.8

  if self.vz > -200 and self.ism:getT() < 0.5 and input.isJumpHit(self.id) then
    audio.playRandom('throw') -- reusing the same sound?
    self:emitDust('rescue_jump')
    self.ism:set('jump')
  end

  self.ism:after(FAIL_FALL_T, 'intro') -- random pos?
end

function mt:onFailFallEnd()
  audio.playRandom('fall_die')
end

--------------- IDLE STATE ---------------

function mt:onIdleStart()
end

function mt:onIdleUpdate(dt)
  -- if self.vz < 0 then
    self.vz = 0
  -- end
  self:playAnim('idle')
  self:updateMovement(dt) -- we can start movement even in idle
  if self:isMoving() then self.ism:set('run') end
  self:tryJump()
  self:tryFailFall()
  self:tryThrow()
end

--------------- STOMP STATE ---------------
function mt:onStompStart(dt)
  self.vx, self.vy, self.vz = 0, 0, 420
end
function mt:onStompUpdate(dt)
  self.z = self.z + self.vz * dt
  self.vz = self.vz - STOMP_GRAVITY * dt
  if self.z <= 0 then
    self.z = 0
    self.ism:set('squash')
  end
end

function mt:onStompEnd()
  if self:hasGroundUnderneath() then
    self:emitDust('stomp')
    audio.playRandom('stomp')
    signal.emit('stomp', self.id, 0.5, self:getOrigin()) -- stomp is not that powerful
  else
    self:tryFailFall()
  end
end

--------------- FALL STATE ---------------
function mt:onFallStart(dt)
  self.vz = 0
end

function mt:onFallUpdate(dt)
  self:playAnim('fall')

  self.z = self.z + self.vz * dt
  self.vz = self.vz - FALL_GRAVITY * dt

  if self.z <= 0 then
    self.z = 0
    self.ism:set('squash')
  end
end

function mt:onFallEnd()
  self:onJumpEnd()
end

--------------- JUMP STATE ---------------
function mt:onJumpStart(dt)
  self.z = 0.01 -- just so it's above 0
  self.vz = INIT_JUMP_V
  self:emitDust('jump')
  audio.playRandom('jump')
end

function mt:onJumpUpdate(dt)
  self:playAnim(self.vz > 0 and 'rise' or 'fall')

  local g = self.vz < 0 and FALL_GRAVITY or (input.isJumpHold(self.id) and RISE_GRAVITY or RISE_LOW_GRAVITY)

  self.z = self.z + self.vz * dt
  self.vz = self.vz - g * dt -- "-" beacuse we want to be falling

  self:updateMovement(dt)

  if self.z <= 0 then
    self.z = 0
    self.ism:set('squash')
  elseif self.z > 10 and input.isJumpHit(self.id) then
    self.ism:set('stomp')
  end

  self:tryThrow()
end

function mt:onJumpEnd()
  if self.z > 0 then return end -- otherwise we changed to stomp
  if self:hasGroundUnderneath() then
    self:emitDust('fall')
    audio.playRandom('fall')
  else
    self:tryFailFall()
  end
end

--------------- SQUASH STATE ---------------
function mt:onSquashUpdate(dt)
  self.ism:after(SQUASH_TIME, self:isMoving() and 'run' or 'idle')
  self:playAnim(self.ism:getT() >= (SQUASH_TIME-0.01) and 'idle' or 'jump_squash') -- rounding hack

  self:updateMovement(dt)
  self:tryFailFall()
end

--------------- RUN STATE ---------------
function mt:onRunUpdate(dt)
  -- if self.vz < 0 then
    self.vz = 0
  -- end
  self:playAnim('run')
  self:updateMovement(dt)
  if not self:isMoving() then self.ism:set('idle') end
  self:tryJump()
  self:tryFailFall()
  self:tryThrow()
end

--------------- THROW STATE ---------------

function mt:onThrowUpdate(dt)
  self:updateMovement(dt)
  self:tryFailFall()
  self.animset:play('throw')
  if self.ism:getT() > 0.1 then
    if self.z > 0.01 then
      self.ism:set('fall')
    else
      self.ism:set('idle')
    end
  end
end

--------------- COMMON ---------------

function mt:canBeStomped()
  if self.ism:is('intro') or self.ism:is('death') or self.ism:is('fail_fall') then
    return false
  end
  return self.z == 0
end

local _carry_anims = {}
function mt:playAnim(name)
  if self.payload then
    name = _carry_anims[name] or (name .. '_carry')
    _carry_anims[name] = name
    self.animset:play(name)
  else
    self.animset:play(name)
  end
end

function mt:tryFailFall()
  if self.z <= 0 and not self:hasGroundUnderneath() then
    self.ism:set('fail_fall')
  end
end

function mt:tryThrow()
  if self.payload and input.isThrowHit(self.id) then
    self.ism:set('throw')
    audio.playRandom('throw')
    local ox, oy = self:getOrigin()
    local ax_x, ax_y = input.getAxes(self.id) -- this should make about_0 -> 0 (analog controlers)
    if ax_x == 0 and ax_y == 0 and not self:isJumping() then
      self.payload:put(ox, oy, self)
    else
      self.payload:throw(ox, oy, self.z+14, self.dir_x, self.dir_y, self.vx, self.vy)
    end
    self.payload = nil
  end
end

function mt:tryJump()
  if input.isJumpHit(self.id) and self:hasGroundUnderneath() then
    self.ism:set('jump')
  end
end

function mt:isMoving()
  return math.abs(self.vx) > 1.5 or math.abs(self.vy) > 1.5
end

function mt:push(vx, vy, vz)
  self.push_vx = self.push_vx + vx
  self.push_vy = self.push_vy + vy

  if vz > 0 then
    self.ism:set('jump') -- unexpected but it gives interesting results, also: can we have rocketjumps?
  end
end

function mt:onCollision(other, col)
  if other.is_box then
    if self:isJumping() then return end
    if not self.payload then
      signal.emit('player:picks:box', self, other)
      -- map is supposed to give player payload at this point
    else
      -- push
      local px, py = self:getOrigin()
      local ox, oy = other:getOrigin()
      local dx, dy = (px-ox), (py-oy)
      local ndx, ndy = vec.normalize(dx, dy)
      local dist = vec.len(dx, dy)
      local len = math.max(50 - dist, 40)
      self:push(ndx*len, ndy*len, 0)
    end

  elseif other.is_player then
    if not self:isJumping() and not other:isJumping() then
      local px, py = self:getOrigin()
      local ox, oy = other:getOrigin()
      local dx, dy = (px-ox), (py-oy)
      local ndx, ndy = vec.normalize(dx, dy)

      other:push(ndx*-40, ndy*-40, 0)
    end

  elseif other.is_bomb then
    if self:isJumping() then return end
    if not other:canBePickedBy(self) then return end

    if self.payload == nil then
      self.payload = other
      other:bePicked()
    else
      local px, py = self:getOrigin()
      local ox, oy = other:getOrigin()
      local dx, dy = (px-ox), (py-oy)
      local ndx, ndy = vec.normalize(dx, dy)
      self:push(ndx*50, ndy*50, 0)
      other:push(ndx*-100, ndy*-100, 100)
    end

  end
  -- TOOD
  -- if other player => push them
  -- if pickable
  --  if payload => push self away
  --  else => pick
end

function mt:updateMovement(dt)
  if self.is_locked then return end

  local target_x = self.x + self.vx * dt + self.push_vx * dt
  self.vx = self.vx + self.ax * dt
  self.vx = self.vx * 0.85 -- ?

  local target_y = self.y + self.vy * dt + self.push_vy * dt
  self.vy = self.vy + self.ay * dt
  self.vy = self.vy * 0.85 -- ?

  self.push_vx = self.push_vx * 0.7
  self.push_vy = self.push_vy * 0.7

  local new_x, new_y, cols, len = game.world:move(self, target_x, target_y, filters.moveStd)

  for _, col in ipairs(cols) do
    self:onCollision(col.other, col)
  end

  self.x, self.y = new_x, new_y

  local dx, dy = input.getAxes(self.id)
  self.ax, self.ay = dx * ACCELERATION, dy * ACCELERATION
  self:setDir(dx, dy)
end

function mt:hasGroundUnderneath()
  -- OPTIMIZE: per-frame cache?
  local rx, ry, rw, rh = self.x+3, self.y+3, 2, 1 -- very small so it's not too easy to fall
  local _, len = game.world:queryRect(rx,ry,rw,rh, filters.queryTile)
  return len > 0
end

function mt:emitDust(mode)
  local ox, oy = self:getOrigin()
  player_dust[mode](self.dust_psystem, ox, oy)
end

function mt:setDir(dir_x, dir_y)
  if dir_x == 0 and dir_y == 0 then return end -- there always must be some dir!
  self.anim_dir_x = (dir_x == 0) and self.anim_dir_x or (dir_x > 0 and 1 or -1)
  self.dir_x, self.dir_y = vec.normalize(dir_x, dir_y)
end

function mt:getY()
  -- this function sucks so much
  return
    self.y + 5
    + ((self.z or 0) >= 0 and 16 or 0) -- visual bug hack (when falling & touching tile from side)
end

function mt:getOrigin()
  return self.x + 4.5, self.y + 2
end

function mt:getRect()
  return self.x, self.y, 9, 3
end

function mt:isJumping()
  return self.z > 0.001
end

function mt:getColor()
  return colors.idToColor(self.id)
end

function mt:lock()
  self.is_locked = true
end

function mt:unlock()
  self.is_locked = false
end

function mt:kill()
  -- when exiting the map
  self:lock()
  self.ism:set('death')
end

--------------- SETUP ---------------

function Player.new(id, x, y)
  local p = setmetatable({x=x, y=y, id=id}, mt)
  p:init(id, x, y)
  return p
end

function Player.setup(_game)
  game = _game
  game.register('Player', Player)

  local squash_t = SQUASH_TIME / #assets.quadsGroup('player_jump_squash')
  newPlayerAnimset = animset.define({
    run =  { 'player_run',  0.10 },
    idle = { 'player_idle', 0.16 },
    rise = { 'player_rise', 0.10 },
    fall = { 'player_fall', 0.10 },
    jump_squash = { 'player_jump_squash', squash_t },
    throw = { 'player_throw', 0.10 },

    run_carry = { 'player_run_carry',  0.10 },
    idle_carry = { 'player_idle_carry', 0.16 },
    rise_carry = { 'player_rise_carry', 0.10 },
    fall_carry = { 'player_fall_carry', 0.10 },
    jump_squash_carry = { 'player_jump_squash_carry', squash_t },
  })
  Player.payload_offset = { -- sucks to have it hardcoded but there won't be much animations anyway
    default = { 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2 },
    idle_carry = { 6, 5, 5, 6 },
 -- run_carry = { 3, 2, 4, 5, 3, 2, 4, 5 },
    run_carry = { 5, 3, 2, 4, 5, 3, 2, 4 },
  }
  Player.payload_offset.run = Player.payload_offset.run_carry -- avoid error on transition
  Player.payload_offset.idle = Player.payload_offset.idle_carry -- avoid error on transition

end

return Player
