local bump = require('lib.bump')
local signal = require('lib.signal')
local cam = require('cam')
local map = require('map')
local filters = require('utils.filters')
local audio = require('utils.audio')
local later = require('utils.later')
local cmd = require('utils.cmd')
local audio = require('utils.audio')
local vec = require('lib.vector-light')
local flash = require('utils.flash')
local options = require('options')

local game = {
  time_factor = 1,
  entity_classes = {},
  debug_info = false
}


local GAME_W = 427
local GAME_H = 240

local STATES = {
  menu = require('states.menu'),
  play = require('states.play'),
  show_points = require('states.show_points'),
  join_state = require('states.join_state'),
  tutorial = require('states.tutorial')
}

function game.setup()
  -- :recreateCanvas
  game.canvas = love.graphics.newCanvas(1280, 720)
  game.canvas:setFilter('nearest', 'nearest')

  game.input = require('input')
  map.setup(game)
  game.map = map
  cmd.setup(game)
  cam.setup()
  flash.setup()
  audio.setup()
  game.font = love.graphics.newFont('assets/press_start/prstartk.ttf', 8)
  game.font:setFilter('nearest', 'nearest')
  love.graphics.setFont(game.font)

  game.bots = require('utils.bots')
  game.bots.setup(game, game.map)
  -- game.bots.add() -- todo

  for k, v in pairs(STATES) do v.setup(game) end

  -- signal.register('stomp', require('events.on_stomp')(game))

  game.sh = love.graphics.newShader('shaders/std_shader.glsl') -- let us blink
end

function game.register(name, class)
  game.entity_classes[name] = class
end

function game.spawn(class_name, ...)
  return game.entity_classes[class_name].new(...)
end

function game.newMap(name)
  game.world = bump.newWorld(64)
  game.to_be_removed = {}
  map.load(name)
end

function game.unloadMap()
  map.unload()
end

function game.add(e)
  game.world:add(e, e:getRect())
end

function game.safeAdd(e)
  if game.world:hasItem(e) then return end
  game.world:add(e, e:getRect())
end

function game.remove(e)
  game.to_be_removed[e] = true
end

function game.cleanToBeRemoved()
  for e in pairs(game.to_be_removed) do
    game.world:remove(e)
    game.to_be_removed[e] = nil
  end
end

function game.setState(name)
  if game.curr_state and game.curr_state.onClose then
    game.curr_state.onClose()
  end
  game.curr_state = STATES[name]
  game.curr_state_name = name
  STATES[name].init()
  collectgarbage()
end

function game.keypressed(k)
  if game.curr_state.keypressed then
    game.curr_state.keypressed(k)
  end
end

function game.toggleBlink(val)
  game.sh:send('blink', val and 10000 or 1)
end

function game.getScale()
  if options.video.scale == 'full' then
    local gw, gh = game.getDimensions()
    local ww, wh = love.graphics.getDimensions()
    local sw, sh = gw/ww, gh/wh
    return 1/math.max(sw, sh)
  else
    return options.video.scale
  end
end

function game.getScaleOffset()
  local scale = game.getScale()
  local gw, gh = game.getDimensions()
  local ww, wh = love.graphics.getDimensions()
  local sgw, sgh = gw*scale, gh*scale

  return (sgw-ww)/-2, (sgh-wh)/-2
end

function game.getDimensions()
  return GAME_W, GAME_H
end

function game.draw()
  love.graphics.setShader(game.sh)
  love.graphics.clear()

  cam.push()
  love.graphics.translate(game.getScaleOffset())
  love.graphics.setColor(255, 255, 255)
  love.graphics.scale(game.getScale())
  game.curr_state.draw()

  love.graphics.setShader()
  if game.curr_state.hudDraw then game.curr_state.hudDraw() end
  cam.pop()

  cmd.draw()

  game.bots.draw()
  love.graphics.setColor(255, 255, 255)

  -- MAX_DRAW_T = math.max((love.timer.getTime() - T), MAX_DRAW_T or 0)
  -- print("D", (love.timer.getTime() - T)*1000, MAX_DRAW_T)
end

function game.find(predicate)
  local t = type(predicate)
  if t == 'string' then
    return game.world:queryRect(-10000, -10000, 20000, 20000, function(item)
      return item[predicate]
    end)
  elseif t == 'function' then
    return game.world:queryRect(-10000, -10000, 20000, 20000, predicate)
  else
    error('wrong predicate type, expected string or function, got ' .. t .. ': ' .. tostring(predicate))
  end
end

function game.setTimeFactor(tf)
  game.time_factor = tf
end

function game.update(irl_dt)
  -- local T = love.timer.getTime()

  local fake_time_factor = nil
  -- if love.keyboard.isDown('space') then -- debug
  --   fake_time_factor = 5
  -- end

  local dt = irl_dt * (fake_time_factor or game.time_factor or 1)

  if not game.is_stopped then
    game.input.update(dt)
    game.curr_state.update(dt)
    map.update(dt) -- map module must know if anything is loaded or not (?)
  end

  game.bots.update(dt)

  if
    (not cmd.isEnabled()) and
    (love.keyboard.isDown('lshift', 'rshift') and
      (love.keyboard.isDown(';') or love.keyboard.isDown(','))
    ) then
    cmd.enable()
  end

  later.update(dt)
  cam.update(dt)


  -- perf
  -- if (love.timer.getTime() - T) > (MAX_T or 0) then
  --   MAX_T = (love.timer.getTime() - T)
  -- end
  --
  -- if (love.timer.getTime() - T) * 1000 > 1 then
  --   print("U:", (love.timer.getTime() - T) * 1000, MAX_T*1000)
  -- end

  -- if game.world then
  --   local _, len = game.find('is_explosion')
  --   print("ELEMS COUNT: ", len)
  -- end
end

return game
