local COUNTDOWN = 60 -- secs

local signal = require('lib.signal')
local vec = require('lib.vector-light')
local later = require('utils.later')
local audio = require('utils.audio')
local filters = require('utils.filters')

local _maps = {
  basic = require('maps.basic')
}

local game = nil
local map = {}

local function _affectTiles(x, y, power, radius)
  local items, len = game.world:queryRect(x-30, y-30, 60, 60, filters.queryTile)

  for _, tile in ipairs(items) do
    local cx, cy = tile:getCenter()
    local dist = vec.len(cx-x, cy-y)

    -- shake
    local sh = (1 - dist / 100)
    if sh > 0 then
      tile:shake(sh*5, sh*2)
    end

    -- damage
    local dmg = (4 - math.floor(dist/6)) * 0.2 -- actually stomping it's preatty weak to the tiles
    if dmg > 0 then
      tile:damage(dmg * power)
    end
  end
end

local function _affectOtherPlayers(x, y, power, radius, player_id)
  if power > 0.01 and player_id then
    local items, len = game.world:queryRect(x-radius, y-radius, radius*2, radius*2, filters.queryPlayer)

    for _, p in ipairs(items) do
      -- you won't stomp on yourself, pass 99 to player_id if you want
      if p.id ~= player_id and p:canBeStomped() then

        local sub_x, sub_y = vec.sub(p.x, p.y, x, y)
        if sub_x == 0 then sub_x = 0.1 end
        if sub_y == 0 then sub_y = 0.1 end
        local dist = vec.len(sub_x, sub_y)

        if dist < radius then
          local px, py = vec.normalize(sub_x, sub_y)
          p:push(px*300*power, py*300*power, 200)
          p.jump_z = -0.01
        end

        audio.playRandom('head_fall')
      end
    end
  end
end

local function _affectBombs(x, y)
  local radius = 60
  local items, len = game.world:queryRect(x-radius, y-radius, radius*2, radius*2, filters.queryBomb)

  for _, b in ipairs(items) do
    local bx, by = b:getOrigin()
    local sub_x, sub_y = vec.sub(bx, by, x, y)
    if vec.len(sub_x, sub_y) < radius then
      local px, py = vec.normalize(sub_x, sub_y)
      b:push(px*75, py*75, 150)
    end
  end
end

local function _onStomp(player_id, power, x, y)
  if power < 0.1 then return end
  local radius = 40
  _affectTiles(x, y, 0.1, radius)
  _affectOtherPlayers(x, y, 2, radius, player_id)
  _affectBombs(x, y)
end

local function _onExplosion(x, y) -- this all should go to explosion!
  local radius = 40
  local power = 1.5
  _affectTiles(x, y, power, radius)
  _affectOtherPlayers(x, y, 4, radius, 99) -- 99 fake player id
end

function map.setup(_game)
  game = _game

  signal.register('stomp', _onStomp)
  signal.register('explosion', _onExplosion)

  signal.register('box:exploded', function(player, box)
    map.current.onBoxRemoved(map) -- spawn a new one!
  end)

  signal.register('player:picks:box', function(player, box)
    player.payload = map.current.getPayload()
    box:onBePicked(player)
    map.current.onBoxRemoved(map)
  end)

  signal.register('recalc_points', function()
    if map.countdown then return end

    local _, len = game.find(filters.queryPointTile)
    local max = 24 * map.numberOfPlayers
    if len < max / 2 then
      map.countdown = COUNTDOWN
    end
  end)
end

function map.killAllPlayers()
  audio.playRandom('tile_fall') -- reuse
  for i, p in ipairs(map.players) do
    p:kill()
  end
end

local _ii_uid = 0
function map.load(name)
  print('load map: ', name)

  map._spot_count = 0
  map.current = _maps.basic
  map.current.build(game, map)
  map.t = 0
  _ii_uid = _ii_uid + 1
  map.uid = name .. '###' .. _ii_uid
  map.release_players_flag = false
  map.players = {}
  map.countdown = nil -- 10 -- debug

  local _, init_size = game.find(filters.queryTile)
  map.init_size = init_size

  local _current_uid = map.uid
  local prt = map.current.getPlayersReleaseTime()

  for id=1,map.numberOfPlayers do
    later.after(map.current.getPlayerDelay(id), function()
      if not map.uid or map.uid ~= _current_uid then return end
      local p = game.spawn('Player', id, 0, 0)
      p:setDir(map.current.getPlayerDir(id))
      p:lock()
      map.players[id] = p
      game.add(p)
    end)
  end

  later.after(prt - 0.5 - 0.4, function()
    if not map.uid or map.uid ~= _current_uid then return end
    signal.emit('flash')
  end)

  later.after(prt, function()
    if not map.uid or map.uid ~= _current_uid then return end
    for k, p in pairs(map.players) do p:unlock() end
  end)

  -- TODO: depends on number of players!
  local spawnBox = function(pos)
    if not map.uid or map.uid ~= _current_uid then return end
    local bx, by = map.current.getInitBoxPosition(pos)
    local b = game.spawn('Box', bx, by)
    game.add(b)
  end

  later.after(prt - 3.4, function() spawnBox(1) end)
  later.after(prt - 3.3, function() spawnBox(2) end)
  later.after(prt - 3.2, function() spawnBox(3) end)
  later.after(prt - 3.1, function() spawnBox(4) end)
end

function map.unload(name)
  if map.current and map.current.onClose then map.current.onClose() end
  map.current = nil
  map.uid = nil
end

function map.update(dt)
  if not map.current then return end

  map.t = map.t + dt
  -- print("map.box_count", map.box_count)
end

function map.drawCountdown(game)
  if not map.countdown then return end
  local c = math.max(0, map.countdown)
  local t = ('%5.2f'):format(c):gsub(' ', '0'):gsub('%.', ':')
  local w, _ = game.getDimensions()
  love.graphics.printf(t, 0, 10, w/2, 'center', 0, 2, 2)
end

function map.getSafeSpot()
  local tiles, len = game.world:queryRect(-10000, -10000, 20000, 20000, filters.querySafeTile)

  local n = love.math.random(1, len)

  local tile = nil

  for i=1,len do
    local ii = (i + n) % len + 1

    local tx, ty, tw, th = tiles[ii]:getRect()
    local _, len = game.world:queryRect(tx, ty, tw, th, filters.queryBox)

    if len == 0 then
      tile = tiles[ii]
      break
    end
  end

  if not tile then
    signal.emit('no_safe_tile') -- todo: handle?
    return nil
  end

  return tile.x, tile.y
end

function map.calcPoints()
  local _, p1_len = game.find(function(item) return item.is_tile and item.player_id == 1 and item.hp > 0 end)
  local _, p2_len = game.find(function(item) return item.is_tile and item.player_id == 2 and item.hp > 0 end)
  local _, p3_len = game.find(function(item) return item.is_tile and item.player_id == 3 and item.hp > 0 end)
  local _, p4_len = game.find(function(item) return item.is_tile and item.player_id == 4 and item.hp > 0 end)

  return p1_len, p2_len, p3_len, p4_len
end

function map.getSafeSpotForPlayer(p)
  map._spot_count = map._spot_count + 1
  local x, y = map.current.getSafeSpotForPlayer(map._spot_count, p)
  if not x then
    x, y = map.getSafeSpot()
  end

  if not x then
    return nil -- no safe spot
  end

  return x+5, y+7
end


return map
