local map = require('map')
local cam = require('cam')
local signal = require('lib.signal')
local audio = require('utils.audio')
local later = require('utils.later')
local points = require('utils.points')
local colors = require('utils.colors')
local input = require('input')
local flash = require('utils.flash')
local drawFade = require('utils.fade')
local menu = require('utils.menu')

local game = nil
local play_state = {}

local function fadeOutTo(target)
  play_state.fade = -1
  later.after(2.5, function() game.setState(target) end)
end

play_state.pause_menu = menu.new({
  { text='RESUME', exit=true },
  { text='QUIT',
    submenu = {
      { text = 'YES', cb=function()
        fadeOutTo('menu')
        play_state.pause_menu:finalConfirm();
      end },
      { text = 'NO', exit=true },
      title = 'Are you sure you want to quit?',
      on_menu_buton='exit',
      default = 'NO',
      fade = 0.6,
    }
  },
  fade = 0.6,
  on_menu_buton = 'exit',
  default = 'RESUME'
})

local function _ySort(a, b)
  local as, bs = a:getY(), b:getY()
  if as == bs then
    return a.x < b.x
  end
  return as < bs
end

function play_state.setup(_game)
  game = _game
end

function play_state.init()
  play_state.t = 0
  game.newMap('basic')
  audio.playMusic('play_music')
  play_state.fade = nil
  play_state.pause_menu:disable()
end

function play_state.draw()
  if play_state.t == 0 then return end -- skipping first frame to avoid glitches
  local items = game.world:getItems()
  table.sort(items, _ySort)
  for _, e in ipairs(items) do if e.draw then e:draw(dt) end end
  for _, e in ipairs(items) do if e.lateDraw then e:lateDraw(dt) end end
end

function play_state.hudDraw()
  map.drawCountdown(game)
  flash.draw()
  play_state.pause_menu:draw()

  if play_state.fade then
    local f = math.max(0, math.min(play_state.fade, 1))
    drawFade(f)
  elseif play_state.t < 1 then
    drawFade(1 - play_state.t/1)
  end
end

function play_state.shouldTriggerKillAllPlayers()
  return map.countdown and map.countdown <= 0 and not play_state.fade
end

function play_state.update(dt)
  play_state.t = play_state.t + dt

  if not play_state.pause_menu:isEnabled() then
    if map.countdown then map.countdown = map.countdown - dt end -- this should be definitely not here!

    flash.update(dt)

    for e in pairs(game.world.rects) do -- don't care about the order here
      if e.update then e:update(dt) end
    end

    game.cleanToBeRemoved()

    if play_state.shouldTriggerKillAllPlayers() then
      points.add(map.calcPoints())
      map.killAllPlayers()
      fadeOutTo('show_points')
    end
  end

  if play_state.fade then
    play_state.fade = play_state.fade + dt
  end

  play_state.pause_menu:update(dt)
  -- print(input.isMenuHit(1))
  if input.isMenuHit(1) then play_state.pause_menu:enable() end
end



return play_state
