local signal = require('lib.signal')
local filters = require('utils.filters')
local map = require('map')
local cam = require('cam')

local game = nil
local is_enabled = false
local cmd = {}
local before = ''
local after = ''
local history_cursor = 1
local output = ''
local print_err = nil
local _env = {} -- all executions share the same env

local history = {}

function cmd.setup(_game)
  game = _game
  signal.register('textinput', function(t)
    if not cmd.isEnabled() then return end
    before = before .. t
  end)

  signal.register('keypressed', function(k)
    if not cmd.isEnabled() then return end

    local ctrl = love.keyboard.isDown('lctrl', 'rctrl')
    if k == 'backspace' then before = before:sub(1, -2)
    elseif k == 'left' then after = before:sub(-1,-1) .. after; before = before:sub(1, -2)
    elseif k == 'right' then before = before .. after:sub(1,1); after = after:sub(2,-1)
    elseif k == 'a' and ctrl then after = before .. after; before = ''
    elseif k == 'e' and ctrl then before = before .. after; after = ''
    elseif k == 'u' and ctrl then before = ''
    elseif k == 'k' and ctrl then after = ''
    elseif k == 'up' then
      before, after = (history[history_cursor] or ''), ''
      history_cursor = math.max(history_cursor - 1, 1)
    elseif k == 'down' then
      before, after = (history[history_cursor] or ''), ''
      history_cursor = math.min(history_cursor + 1, #history)
    elseif k == 'escape' then
      cmd.disable()
    elseif k == 'return' then
      cmd.exec()
    end
  end)
end

function _print(...)
  print_err = false
  local a = {...}
  local out = {}
  for i, v in ipairs(a) do
    out[i] = tostring(v)
  end
  local str = table.concat(out, '   ')
  output = str
end

function _err(...)
  _print(...)
  print_err = true
end

function cmd.getEnv()
  local players
  if game.world then
    players = game.world:queryRect(-10000, -10000, 20000, 20000, filters.queryPlayer)
  end

  if players then
    table.sort(players, function(a,b) return a.id < b.id end)
  end

  _env.print = _print
  _env.game = game
  _env.world = game.world
  _env.map = map
  _env.math = math
  _env.tostring = tostring
  _env.players = players
  _env.pairs = pairs
  _env.ipairs = ipairs
  _env.getmetatable = getmetatable
  _env.setmetatable = setmetatable
  _env.cam = cam

  return _env
end

function cmd.exec()
  local line = before .. after

  before, after = '', ''
  local hlen = #history
  if history[hlen] ~= line then
    table.insert(history, line)
    if hlen > 512 then
      table.remove(history, 1)
    end
  end
  history_cursor = #history

  -- useful
  if line:sub(1,1) == ',' then
    line = 'print( tostring( ' .. line:sub(2,-1) .. ' ) )'
  end

  local lerr = nil
  local ok, perr = pcall(function()
    local chunk, lerr_inner = load(line, nil, 'bt', cmd.getEnv())
    if chunk then chunk()
    else lerr = lerr_inner end
  end)

  if not ok or lerr then
    _err(tostring(perr or '') .. tostring(lerr or ''))
  end

end

function cmd.isEnabled()
  return is_enabled
end

function cmd.disable()
  after, before, output = '', '', ''
  is_enabled = false
  game.is_stopped = false
end
function cmd.enable()
  after, before, output = '', '', ''
  is_enabled = true
  game.is_stopped = true
end

function cmd.debugger(options)
  cmd.enable()
end

function cmd.draw()
  if not is_enabled then return end
  love.graphics.origin()
  love.graphics.scale(2, 2)
  local h = love.graphics.getHeight() / 2
  local f = love.graphics.getFont()

  if print_err then
    love.graphics.setColor(200, 0, 0)
  else
    love.graphics.setColor(255, 255, 255)
  end

  local _, lines_count = output:gsub('\n', '')
  love.graphics.print(output, 25, h-40-f:getHeight()*lines_count)

  love.graphics.setColor(255, 255, 255)

  love.graphics.print('> ' .. before .. after, 10, h-20)

  local cx = f:getWidth('> ' .. before)
  love.graphics.rectangle('fill', 10+cx, h-25, 1, 15)
  love.graphics.origin()
end


--
-- if love.keyboard.isDown('y') then
--   local line = nil
--   local players = game.world:queryRect(-10000, -10000, 20000, 20000, filters.queryPlayer)
--   while line ~= 'exit' do
--     io.write('LOVE> ')
--     line = io.read('*l')
--     local chunk, err = load(line, nil, 'bt', { game=game, print=print, players=players, math=math })
--     if chunk then chunk() else io.stderr:write(err) end
--   end
-- end

return cmd
